from protlib import CStruct, CInt, CArray

class ResponseJobErrorDescription(CStruct):
    message_length = CInt()
    source_code = CInt()
    source_name_length = CInt()
    error_code = CInt()
    info_elements_length =CInt()
    info_elements = CArray(length="info_elements_length", ctype=CInt)


from protlib import CStruct, CInt, CArray
from .response_job_error_description import ResponseJobErrorDescription
from .response_job_error_data import ResponseJobErrorData

class ResponseJobErrors(CStruct):
    length = CInt()
    count = CInt()
    dummy = CInt()
    description = CArray(length="count", ctype=ResponseJobErrorDescription.get_type())
    data = CArray(length="count", ctype=ResponseJobErrorData.get_type())


from .client import Client
import os
import uuid
import zeep, zeep.exceptions
from typing import List

from .result import Result
from .job import Job
from .param import Param
from .const import ParamTypes
from .result_file import ResultFile, ResultFileType


class SoapClient(Client):
    def __init__(self, endpoint:str, appname:str, username:str, password:str):
        self.endpoint = endpoint
        self.wsdlUri = endpoint + '/EcmWsMtom?wsdl'
        self.client = zeep.Client(wsdl=self.wsdlUri)
        self.enaioWsTypesFactory = self.client.type_factory('ns0')
        self.enaioWsFactory = self.client.type_factory('ns1')
        self.sessionId = str(uuid.uuid4())
        self.auth = self.enaioWsTypesFactory.Authentication(
            user=username,
            password=password,
            applicationName=appname,
            sessionIdentifier=self.sessionId
        )

    def execute(self, job:Job) -> Result:
        """Send a job to the blue web server (SOAP), execute it and return the response.

        Keyword arguments:
        job -- A previously created Job() object.
        """
        params:List[Param] = []
        file_params:List = []
        for param in job.params:
            if param.type == ParamTypes.INTEGER:
                params.append(self.enaioWsTypesFactory.IntegerParameter(name=param.name, value=param.value))
            elif param.type == ParamTypes.BOOLEAN:
                params.append(self.enaioWsTypesFactory.BooleanParameter(name=param.name, value=param.value))
            elif param.type == ParamTypes.DATE_TIME:
                params.append(self.enaioWsTypesFactory.DateTimeParameter(name=param.name, value=param.value))
            elif param.type == ParamTypes.DOUBLE:
                params.append(self.enaioWsTypesFactory.DoubleParameter(name=param.name, value=param.value))
            elif param.type == ParamTypes.STRING:
                params.append(self.enaioWsTypesFactory.StringParameter(name=param.name, value=param.value))
            elif param.type == ParamTypes.BASE64:
                if isinstance(param.value, (bytearray, bytes)):
                    params.append(self.enaioWsTypesFactory.Base64Parameter(name=param.name, value=param.value))
                else: 
                    params.append(self.enaioWsTypesFactory.Base64Parameter(name=param.name, value=bytearray(param.value, "UTF-8")))
            elif param.type == ParamTypes.DB:
                raise Exception('parameter type DB is not supported.')
            else:
                raise Exception('Unknown parameter type')

        for f in job.files:
            reader = open(f, 'rb')
            file_param = self.enaioWsTypesFactory.FileParameter(
                fileName=os.path.basename(f),
                content=self.enaioWsTypesFactory.Content(
                    attachment=reader.read())
            )
            reader.close()
            file_params.append(file_param)

        error_message = None
        result_return_code = -1
        try:
            native_result = self.client.service.execute(
                authentication=self.auth,
                job=self.enaioWsTypesFactory.Job(
                    name=job.name,
                    parameter=params,
                    fileParameter=file_params
                ),
                property=[]
            )
            result_return_code = native_result.job.returnCode  
            result_values = { e['name']: e['value'].decode('UTF-8') if isinstance(e['value'], (bytes, bytearray)) else e['value'] for e in native_result.job.parameter }
            result_files = [ResultFile(ResultFileType.BYTE_ARRAY, file_name=fp.fileName, byte_array=fp.content.attachment) for fp in native_result.job.fileParameter]
            return Result(result_values, result_files, result_return_code, error_message)
        except zeep.exceptions.Fault as ft:
            if ft.message in [ 
                'Incorrect user name or password!',         # 9.00
                'Account check for user [unknown] failed.'  # 9.10
            ]:     
                raise PermissionError(ft.message)
            else:
                error_message = str(ft)
                return Result(return_code=result_return_code, error_message=error_message)            
        except Exception as ex:
            error_message = str(ex)
            return Result(return_code=result_return_code, error_message=error_message)


import socket
import queue
import random
import logging
from random import randint
from threading import Lock

from .result import Result
from .job import Job
from .tcp_client_classes.job_caller import JobCaller
from .tcp_client import TcpClient


class TcpPoolClient(TcpClient):
    def __init__(self, connection_string:str, appname:str, username:str, password:str, use_ssl:bool=True, file_cache_byte_limit:int=33554432, pool_size:int=10):
        servers = []
        for serverString in connection_string.split("#"):
            serverParts = serverString.split(":")
            if len(serverParts) != 3:
                raise Exception(
                    f"Connection String invalid, server '{connection_string}' must be formatted as hostname:port:weight. For example localhost:4000:1")
            servers.append({
                "hostname": serverParts[0],
                "port": int(serverParts[1]),
                "weight": int(serverParts[2])
            })

        self.servers = servers
        self.appname = appname
        self.username = username
        self.password = password
        self.use_ssl = use_ssl
        self.file_cache_byte_limit = file_cache_byte_limit
        self._pool_size = pool_size
        self._pool_available = queue.Queue()
        self._pool_in_use = 0
        self.lock = Lock()

    def __del__(self):
        pass

    def execute(self, job: Job) -> Result:
        job_caller = self._borrow()
        try:
            result = job_caller.execute(job)
            result.client_infos = {
                'hostname': job_caller.hostname, 'port': job_caller.port}
            self._return(job_caller)
            return result
        except ConnectionAbortedError as e:
            # fetch connection closed exceptions and try to reconnect and execute again
            logging.warning(e)
            self._invalidate(job_caller)
            return self.execute(job)

        except Exception as ex:
            self._invalidate(job_caller)
            raise ex

    def _borrow(self) -> JobCaller:
        try:
            job_caller = self._pool_available.get_nowait()
            self._pool_in_use += 1
            return job_caller
        except:
            if(self._pool_in_use >= self._pool_size):
                job_caller = self._pool_available.get()
                self._pool_in_use += 1
                return job_caller
            else:
                job_caller = self._create()
                self._pool_in_use += 1
                return job_caller

    def _invalidate(self, job_caller: JobCaller):
        self._pool_in_use -= 1
        try:
            job_caller.close()
        except:
            pass

    def _return(self, job_caller: JobCaller):
        self._pool_in_use -= 1
        self._pool_available.put(job_caller)

    def _create(self) -> JobCaller:

        self.lock.acquire()
        random.shuffle(rand_servers := [*self.servers])

        job_caller = None
        for server in rand_servers:
            try:
                job_caller = JobCaller(
                    server['hostname'], server['port'], self.use_ssl, self.file_cache_byte_limit)
            except ConnectionRefusedError as cre:
                logging.error(cre)

        if job_caller == None:
            self.lock.release()
            raise ConnectionRefusedError('No valid enaio server found')

        self.lock.release()

        session_attach_job = Job('krn.SessionAttach', Flags=0, SessionGUID='')
        session_attach_result = job_caller.execute(session_attach_job)

        session_properties_set_job = Job(
            'krn.SessionPropertiesSet',
            Flags=0,
            Properties='instname;statname;address',
            address=f'{socket.gethostbyname(socket.gethostname())}=dummy',
            instname=self.appname,
            statname=socket.gethostname()
        )
        session_properties_set_result = job_caller.execute(
            session_properties_set_job)

        session_login_job = Job(
            'krn.SessionLogin',
            Flags=0,
            UserName=self.username,
            UserPwd=TcpClient.encrypt_password(self.password)
        )
        session_login_result = job_caller.execute(session_login_job)

        if session_login_result.values['Description'] != None and session_login_result.values['Description'] != '':
            raise PermissionError(f'Login error: {session_login_result.values["Description"]}')

        return job_caller

    def __del__(self):
        try:
            for i in range(0, self._pool_available.qsize()):
                job_caller = self._pool_available.get_nowait()
                job_caller.socket.close()
        except:
            pass



from enum import Enum


class ParamTypes(Enum):
    STRING = 1
    INTEGER = 2
    BOOLEAN = 3
    DOUBLE = 4
    DATE_TIME = 5
    BASE64 = 6
    DB = 9


class SystemFields(Enum):
    OBJECT_ID = 1100
    OBJECT_COUNT = 1101
    OBJECT_FLAGS = 1102
    OBJECT_AVID = 1103
    OBJECT_AVDATE = 1104
    OBJECT_CRID = 1105
    OBJECT_CRDATE = 1106
    OBJECT_TIME = 1107
    OBJECT_MAIN = 1108
    OBJECT_CO = 1109
    OBJECT_MEDDOCID = 1110
    OBJECT_MEDDIAID = 1111
    OBJECT_MEDDOCNA = 1112
    OBJECT_MEDDIANA = 1113
    OBJECT_LINKS = 1114
    OBJECT_VERID = 1115
    OBJECT_LOCKUSER = 1116
    OBJECT_SYSTEMID = 1117
    OBJECT_MODIFYTIME = 1118
    OBJECT_MODIFYUSER = 1119
    OBJECT_FOREIGNID = 1124
    OBJECT_USERGUID = 1125
    OBJECT_DELETED = 1126
    OBJECT_INDEXHISTFLAGS = 1127
    OBJECT_DOCHISTFLAGS = 1128
    OBJECT_OSSD = 1129
    OBJECT_MIMETYPEID = 1900
    OBJECT_FILESIZE = 1902
    OBJECT_RETENTION_PLANNED = 1903
    OBJECT_RETENTION = 1904
    STAMM_ID = 1000
    STAMM_TIME = 1001
    STAMM_LINKS = 1002
    REG_ID = 1120
    REG_STAID = 1121
    REG_PARID = 1122
    SDSTA_ID = 1130
    SDOBJ_ID = 1131
    SDOBJTYPE = 1132
    SDREG_ID = 1133
    SDDEL = 1134 
    SDTIME = 1135
    SDREG_TYPE = 1136
    FOLDERID = 1181
    FOLDERTYPE = 1182
    REGISTERID = 1183
    REGISTERTYPE = 1184
    PARENTREGID = 1185
    PARENTREGTYPE = 1186 
    MDDEL = 1140
    MDTIME = 1141
    MDMAP_ID = 1142
    MDSTA_ID = 1143
    MDOBJ_ID = 1144
    MDOBJTYPE = 1145
    MDMOD = 1146
    MDIN = 1147
    MDOUT = 1148
    MDCOUNT = 1149 


class ImportActions(Enum):
    NONE = "NONE"
    INSERT = "INSERT"
    UPDATE = "UPDATE"
    ERROR = "ERROR"


class MainTypeId(Enum):
    FOLDER = 0
    REGISTER = 99
    DOC_GRAYSCALE = 1
    DOC_BW = 2
    DOC_COLOR = 3
    DOC_WINDOWS = 4
    DOC_MULTIMEDIA = 5
    DOC_MAIL = 6
    DOC_XML = 7
    DOC_CONTAINER = 8
    DOC_TYPELESS_USER = 200
    DOC_TYPELESS_WORKFLOW = 300
    PORTFOLIO = 203
    NOTE = 32767


class SortOrder(Enum):
    NONE = 0
    ASC = 1
    DESC = -1


class QueryOperators(Enum):
    LOWER_THEN = '<'
    LOWER_EQUAL = '<='
    EQUAL = '='
    NOT_EQUAL = '!='
    GREATER_THEN = '>'
    GREATER_EQUAL = '>='
    BETWEEN = 'BETWEEN'
    NOT_BETWEEN = 'NOT BETWEEN'
    IN = 'IN'
    NOT_IN = 'NOT_IN'


class SpecialValues(Enum):
    """Possible string query values for the `<SpecialValue>` tag"""
    COMPUTER_GUID = '#COMPUTER-GUID#'
    COMPUTER_NAME = '#COMPUTER-NAME#'
    COMPUTER_IP = '#COMPUTER-IP#'
    CREATOR = '#ANLEGER#'
    CREATION_DATE = '#ANLEGEDATUM#'
    ARCHIVIST = '#ARCHIVAR#'
    ARCHIVE_DATE = '#ARCHIVIERUNGSDATUM#'
    USER = '#BENUTZER#'
    OWNER = '#BESITZER#'
    DATE = '#DATUM#'


class ObjectSearchFlags(Enum):
    """Possible numeric query values for the system field `OBJEKT_SEARCHFLAGS`"""
    ARCHIVED = 1
    ARCHIVABLE = 2
    NOT_ARCHIVABLE = 4
    WITHOUT_PAGES = 8
    CHECKOUT_BY_ME = 16
    CHECKOUT_BY_OTHER = 32
    IN_REGISTER = 64
    NOT_IN_REGISTER = 128
    EXTERNAL = 256
    LINK = 512
    MULTI_LOCATION = 1024
    HAS_VARIANTS = 2048
    SIGNED_CURRENT_VERSION = 4096
    SIGNED_FORMER_VERSION = 8192


class FieldGroupOperators(Enum):
    OR = 'OR'
    AND = 'AND'


class StoreInCacheByIdConversion(Enum):
    """Possible numeric values for the `Convert` parameter of `std.StoreInCacheById`"""
    NONE = 0
    PDF = 1
    MULTIPAGE_TIFF = 8



from typing import Union
from .const import SortOrder, SystemFields
from XmlElement import XmlElement

class QueryResultField():
    def __init__(
                self, 
                field:Union[str, SystemFields], 
                sort_pos:Union[int, None]=None, 
                sort_order:Union[SortOrder, None]=None
        ):
        """Create a new QueryResultField() object.

        Keyword arguments:
        field -- A internal name as string or a instance of SystemFields. Internal names beeing automatically checked against known system field names.
        sort_pos -- (Optional) int of the sort position in a query, default = None.
        sort_order -- (Optional) SortOrder instance for this field, default = SortOrder.NONE if sort_pos = None else SortOrder.ASC
        """ 

        if isinstance(field, SystemFields):
            self.internal_name = field.name
            self.system = True
        elif field in [ system_field.name for system_field in SystemFields]:
            self.internal_name = field
            self.system = True
        else:
            self.internal_name = field
            self.system = False

        self.sort_pos = sort_pos

        if sort_order == None and sort_pos != None:
            self.sort_order = SortOrder.ASC
        else:
            self.sort_order = sort_order if sort_order else SortOrder.NONE


    def __repr__(self) -> str:
        return  f'{self.internal_name} (system:{1 if self.system else 0}, sort_pos:{self.sort_pos}, sort_order:{self.sort_order.name})'

    
    def to_xml_element(self) -> XmlElement:
        """Render self to XmlElement"""
        x = XmlElement('Field', {'internal_name': self.internal_name})
        if self.system:
            x.attributes['system'] = '1'

        if self.sort_pos and self.sort_pos > 0 and self.sort_order != SortOrder.NONE:
            x.attributes['sortpos'] = str(self.sort_pos)
            x.attributes['sortorder'] = self.sort_order.name
        
        return x


from .client import Client
from comtypes.client import CreateObject
from comtypes import COMError
import os
from typing import List

from .result import Result
from .result_file import ResultFile, ResultFileType
from .job import Job
from .param import Param
from .const import ParamTypes


class ComClient(Client):
    def __init__(self, hostname:str, port:int, username:str, password:str):
        self.hostname = hostname
        self.port = port
        self.username = username
        self.server = CreateObject('OxSvrSpt.Server')
        self.session = self.server.Login(username, password, hostname, str(port))

    def execute(self, job:Job) -> Result:
        """Send a job to the com server (OxSvrSpt.Server), execute it and return the response.

        Keyword arguments:
        job -- A previously created Job() object.
        """
        native_job = self.session.NewJob(job.name)

        for param in job.params:
            if param.type == ParamTypes.INTEGER:
                native_job.InputParameters.AddNewIntegerParameter(param.name, param.value)
            elif param.type == ParamTypes.BOOLEAN:
                native_job.InputParameters.AddNewBooleanParameter(param.name, param.value)
            elif param.type == ParamTypes.DATE_TIME:
                native_job.InputParameters.AddNewDatetimeParameter(param.name, param.value)
            elif param.type == ParamTypes.DOUBLE:
                native_job.InputParameters.AddNewDoubleParameter(param.name, param.value)
            elif param.type == ParamTypes.STRING:
                native_job.InputParameters.AddNewStringParameter(param.name, param.value)
            elif param.type == ParamTypes.BASE64:
                native_job.InputParameters.AddNewXmlParameter(param.name, param.value)
            elif param.type == ParamTypes.DB:
                raise Exception('parameter type DB is not supported.')
            else:
                raise Exception('Unknown parameter type')

        for f in job.files:
            native_job.InputFileParameters.AddExistFile(f)

        error_message = None
        try:
            native_job.Execute()
            
            result_values = {}
            for op in native_job.OutputParameters:
                result_values[op.Name] = op.XML if op.type == 6 else op.Value

            result_files = []
            for ofp in native_job.OutputFileParameters:
                _, file_name = os.path.split(ofp.FileName)
                rf = ResultFile(ResultFileType.FILE_PATH, file_name=file_name, file_path=ofp.FileName)
                ofp.AutoDelete = False
                result_files.append(rf)
            
            result_return_code = native_job.Errors.ResultCode
            return Result(result_values, result_files, result_return_code, error_message)
        except COMError as ex:
            error_message = str(ex.details)
            return Result(return_code=ex.hresult, error_message=error_message)


from typing import List
from .param import Param
from .const import ParamTypes
from datetime import datetime
from XmlElement import XmlElement

class Job():
    def __init__(self, jobname:str, files:List[str]=[], **params):
        """Create a new Job() object.

        Keyword arguments:
        jobname -- String with the a blue jobname, i. e. 'dms.GetResultList'
        files -- (Optional) List of strings with file paths to add to the job.
        **params -- Add arbitrary job input parameters. Uses Param.infer_type() to guess the blue parameter type.
        """
        self.name = jobname
        self.params:List[Param] = []
        self.files = files
        for name, value in params.items():
            self.append(Param.infer_type(name, value))


    def __repr__(self):
        return self.name


    def append(self, param:Param):
        """Appends a job input parameter.
        Keyword arguments:
        param -- Param object.
        """
        self.params.append(param)


    def update(self, param:Param):
        """Updates a job input parameters value and type. Appends the parameter if not allready present.
        Keyword arguments:
        param -- Param object.
        """
        for p in self.params:
            if p.name == param.name:
                p.value = param.value
                p.type = param.type 
                return True
        self.append(param)

    def append_file(self, filepath:str):
        """Appends a job input file parameter.
        Keyword arguments:
        filepath -- String with file path to append.
        """        
        self.files.append(filepath)


    def __repr__(self) -> str:
        return f'Job "{self.name}" ({len(self.files)} files): {self.params}'


from abc import ABC, abstractmethod
from ecmind_blue_client.options import Options
from .query_condition_group import QueryConditionGroup
from .query_condition_field import QueryConditionField
from .query_result_field import QueryResultField
from typing import Iterator, Optional, Any, List, Dict, Union
from XmlElement import XmlElement
from .job import Job
from .result import Result
from .result_file import ResultFile
from .const import *
from .param import Param, ParamTypes
from datetime import datetime
import logging

log = logging.getLogger(__name__)
class Client(ABC):

    @abstractmethod
    def execute(self, job:Job) -> Result:
        """Send a job to the blue server, execute it and return the response.

        Keyword arguments:
        job -- A previously created Job() object.
        """
        raise NotImplemented
    
    def get_object_type_by_id(self, object_id:int) -> int:
        """Helper function: Execute the dms.GetObjectTypeByID job for a given object id and return the objects type id.

        Keyword arguments:
        object_id -- A folder, register or document id.
        """
        job = Job('dms.GetObjectTypeByID', Flags=0, ObjectID=object_id)
        return self.execute(job).values['ObjectType']


    def store_in_cache(self, object_id:int, object_type_id:int=None, checkout:bool=False) -> List[ResultFile]:
        """Helper function: Execute the std.StoreInCache job for a given object to retrieve its files.

        ### Keyword arguments:

        - object_id -- A document id.
        - object_type_id -- (Optional) The documents type id. When not provided, it is retrieve via get_object_type_by_id() first.
        - checkout -- (Optional) When True, change the documents state to checked out on the server.

        ### Warning

        Known bug: Throws an enaio error on archived documents. 
        """
        if object_type_id == None:
            object_type_id = self.get_object_type_by_id(object_id)

        job = Job('std.StoreInCache',
            Flags=1,
            dwObjectID=object_id,
            dwObjectType=object_type_id,
            DocState=(0 if checkout else 1),
            FileCount=0
        )
        result = self.execute(job)
        if result.return_code != 0:
            raise(RuntimeError(f'Received return code {result.return_code}: {result.error_message}'))

        return result.files


    def store_in_cache_by_id(
        self, object_id:int, convert:StoreInCacheByIdConversion=StoreInCacheByIdConversion.NONE) -> List[ResultFile]:
        """Helper function: Execute the std.StoreInCacheById job for a given object to retrieve/convert its files.

        ### Keyword arguments:

        - object_id -- A document id.
        - convert -- (Optional) Convert 
           - main type 1, 2, 3 or 4 to PDF with StoreInCacheByIdConversion.PDF 
           - main type 2 or 3 with StoreInCacheByIdConversion.MULTIPAGE_TIFF
        """

        job = Job('std.StoreInCacheById',
            Flags=1,
            dwObjectID=object_id,
            Convert=convert.value
        )

        result = self.execute(job)
        if result.return_code != 0:
            raise(RuntimeError(f'Received return code {result.return_code}: {result.error_message}'))

        return result.files


    def xml_import(self,
                  object_name:str,
                  search_fields:Dict[str, str],
                  import_fields:Dict[str, str],
                  table_fields:Dict[str, List[Dict[str, Any]]]=None,
                  folder_id:int=None,
                  register_id:int=None,
                  object_id:int=None,
                  options:Union[Options, str]='',
                  action0:ImportActions=ImportActions.INSERT,
                  action1:ImportActions=ImportActions.UPDATE,
                  actionM:ImportActions=ImportActions.ERROR,
                  files:List[str]=[],
                  main_type:Union[MainTypeId, int, None]=None,
                  variant_parent_id:int=None
        ):
        """Helper function: Execute the dms.XMLImport job.

        ### Keyword arguments
        
        - `object_name` -- The internal name of the object type to import.
        - `search_fields` -- Dict of internal field names and values. If one or more objects match all `search_fields`, the `action1` or `actionM` will be used, `action0` otherwise. 
        - `import_fields` -- Dict of internal field names and (new) values.
        - `table_fields` -- Dict of internal table field names and list of new rows as dicts of internal column name and values.
        - `folder_id` -- (Optional) Folder id to import registers or documents into.
        - `register_id` -- (Optional) Register id to import sub-registers or documents into.
        - `object_id` -- (Optional) Objekt id to force an update of this element. 
        - `options` -- (Optional) Semicolon separated string of import options.
        - `action0` -- (Optional) `ImportActions` Enum element defining how to handle imports when the search_fields do not match any pre-existing objects.
        - `action1` -- (Optional) `ImportActions` Enum element defining how to handle imports when the search_fields do match exactly one pre-existing object.
        - `actionM` -- (Optional) `ImportActions` Enum element defining how to handle imports when the search_fields do match more then one pre-existing object.
        - `files` -- (Optional) List of strings containing file path to import into a document object.
        - `main_type` -- (Optional) Set the main type id for document imports or leave empty for default value. Valid ids are `DOC_GRAYSCALE`/`1`, `DOC_BW`/`2`, DOC_COLOR`3`, `DOC_WINDOWS`/`4`, `DOC_MULTIMEDIA`/`5`, `DOC_MAIL`/`6`, `DOC_XML`/`7`, `DOC_CONTAINER`/`8`.
        - `variant_parent_id` -- (Optional) Set the parent id for document variant imports. `search_fields` must be empty when `variant_parent_id` is set.
        """

        object_element = XmlElement('Object', 
            s=[ XmlElement(
            'Search', s=[ search_fields_element := XmlElement('Fields') ]), 
            import_fields_element := XmlElement('Fields') 
        ])

        if main_type and isinstance(main_type, MainTypeId):
            object_element.set('maintype', str(main_type.value))
        elif main_type:
            object_element.set('maintype', str(main_type))

        if variant_parent_id:
            if len(search_fields) > 0:
                raise ValueError('search_fields must be empty when variant_parent_id is set.')
            search_fields = { SystemFields.OBJECT_ID.name: '-1' }
            object_element.set('variantparent_id', str(variant_parent_id))
        
        xml = XmlElement('DMSData', s=[
                XmlElement('Archive', s=[
                    XmlElement('ObjectType', {'internal_name': object_name}, [object_element])
                ])
        ])
        
        folder_id and object_element.set('folder_id', str(folder_id))
        register_id and object_element.set('register_id', str(register_id))
        object_id and object_element.set('object_id', str(object_id))

        for field_internal_name, field_value in search_fields.items():
            field = XmlElement('Field', a={'internal_name': field_internal_name}, t=field_value)
            if field_internal_name in SystemFields._member_names_:
                field.set('system', '1')
            search_fields_element.append(field)

        for field_internal_name, field_value in import_fields.items():
            field = XmlElement('Field', a={'internal_name': field_internal_name}, t=field_value)
            if field_internal_name in SystemFields._member_names_:
                field.set('system', '1')
            import_fields_element.append(field)

        if table_fields != None and len(table_fields):
            object_element.append(table_fields_element := XmlElement('TableFields'))
            for table_field_internal_name, table_field_rows in table_fields.items():
                table_field = XmlElement('TableField', {'internal_name': table_field_internal_name})
                table_fields_element.append(table_field)
                for table_field_row in table_field_rows:
                    table_field.append(XmlElement('Row', s=[ XmlElement(
                        'Field', {'internal_name': table_row_field_internal_name}, t=table_row_field_value) \
                            for table_row_field_internal_name, table_row_field_value in table_field_row.items()
                    ]))

        job = Job('dms.XMLImport', Flags=0, Options=str(options), 
            Action0=action0.value, Action1=action1.value, ActionM=actionM.value, 
            files=files, Encoding='UTF-8', XML=xml
        )

        return self.execute(job)

    def get_object_details(self, object_name:str, object_id:int, system_fields:Optional[List[SystemFields]]=[]) -> dict:
            query_fields = []
            for system_field in system_fields:
                query_field = XmlElement('Field', {'internal_name': system_field.name, 'system': '1'})
                query_fields.append(query_field)

            query_xml = XmlElement('DMSQuery', {}, [
                XmlElement('Archive', {} , [
                    XmlElement('ObjectType', {'internal_name': object_name}, [
                        XmlElement('Fields', {'field_schema': 'ALL'}, query_fields),
                        XmlElement('Conditions', {}, [
                            XmlElement('ConditionObject', {'internal_name': object_name}, [
                                XmlElement('FieldCondition', {'internal_name': SystemFields.OBJECT_ID.name, 'operator': '=', 'system': '1'}, [
                                    XmlElement('Value', t=str(object_id))
                                ])
                            ])
                        ])
                    ])
                ])
            ])

            job = Job('dms.GetResultList', Flags=0, Encoding='UTF-8', RequestType='HOL', XML=query_xml)
            result = self.execute(job)
            if result.return_code != 0:
                raise(Exception(f'Failed with error code: {result.return_code}'))

            if result.values['Count'] == 0:
                return {}

            result_xml = XmlElement.from_string(result.values['XML'])
            obj_xml = result_xml['Archive'][0]['ObjectType'][0]['ObjectList'][0]['Object'][0]

            result = {}
            for field in obj_xml['Fields'][0]['Field']:
                result[field.attributes['internal_name']] = field.text

            table_fields = obj_xml.find('TableFields')
            if table_fields:
                for table_field in table_fields['TableField']:
                    table_field_result = []
                    cols = [ col.attributes['internal_name'] for col in table_field['Columns'][0]['Column'] ]
                    for row in table_field.findall('Row'):
                        row_result = {}
                        for i, col in enumerate(cols):
                            row_result[col] = row.subelements[i].text
                        table_field_result.append(row_result)

                    result[table_field.attributes['internal_name']] = table_field_result

            return result

    def execute_sql(self, sql_command:str) -> Union[List[Dict[str, str]], None]:
        """Helper function: Execute a sql command via ADO.

        Keyword arguments:
        
        - sql_command -- The sql command.

        Returns:
        
        - For SELECT statements: `List[Dict]` -- The list of records with each row as dictionary of column name and string-value
        - For other statements: `None`
        """
        
        job = Job('ado.ExecuteSQL', Flags=0, Command=sql_command.strip())
        result = self.execute(job)
        if result.return_code != 0:
            raise Exception(f'Failed with error code: {result.return_code}')

        if result.files:
            xml_result = XmlElement.from_string(result.files[0].bytes().decode('UTF-8'))
            data = [ row.attributes for row in xml_result['{urn:schemas-microsoft-com:rowset}data'][0]['{#RowsetSchema}row'] ]
            return data
        else:
            return None

    def lol_query(self,
                        object_name:str, 
                        conditions:Union[QueryConditionGroup, List[QueryConditionGroup]]=None,
                        result_fields:Union[str, SystemFields, QueryResultField, List[Union[str, SystemFields, QueryResultField]]]=None,
                        offset:int=0,
                        page_size:int=1000,
                        max_hits:Union[int, None]=None,
                        include_file_info:bool=False,
                        follow_doc_links:bool=False,
                        include_status:bool=False,
                        garbage_mode:bool=False,
                        context_user:str=None
        ) -> Iterator[Dict[str, Any]]:
        """Helper function: Page through dms.GetResultList in LOL format, yielding dicts of result fields for each hit.

        Keyword arguments:
        object_name -- The internal name of the object type to import.
        result_fields -- List of internal field names, SystemField instances or QueryField instances (for sorting). 
        offset -- (Optional) int defining the query offset, default = 0.
        page_size -- (Optional) int defining the query page size, default = 1000.
        max_hits -- (Optional) int|None limiting the querys maximum yield, default = None.
        include_file_info -- (Optional) bool indicating, if file size, file extension and mime type should be added to document results, default = False.
        follow_doc_links -- (Optional) bool indicating, the include_file_info returns the data of referenced files ("green arrows"), default = False.
        include_status -- (Optional) bool indicating, if status field should be added to the result, default = False.
        garbage_mode -- (Optional) bool indicating, if the query searches the recycle bin instead of non-deleted objects, default = False.
        context_user -- (Optional) string to change the job context to another user name.
        """        

        if conditions and isinstance(conditions, list):
            conditions_element = XmlElement('Conditions', {}, [ x.to_xml_element() for x in conditions ])
        elif conditions:
            conditions_element = XmlElement('Conditions', {}, [ conditions.to_xml_element() ])
        else:
            conditions_element = XmlElement('Conditions')

        if result_fields and len(result_fields) > 0:
            fields_element = XmlElement('Fields', {'field_schema': 'DEF'})
            for result_field in result_fields if isinstance(result_fields, list) else [ result_fields ]:
                if isinstance(result_field, QueryResultField):
                    fields_element.append(result_field.to_xml_element())
                else:
                    fields_element.append(QueryResultField(result_field).to_xml_element())

        else:
            fields_element = XmlElement('Fields', {'field_schema': 'ALL'})

        query_xml = XmlElement('DMSQuery', {}, [
            XmlElement('Archive', {} , [
                XmlElement('ObjectType', {'internal_name': object_name}, [
                    fields_element, conditions_element
                ])
            ])
        ])

        job = Job('dms.GetResultList', 
                    Flags=0, 
                    Encoding='UTF-8', 
                    RequestType='LOL',
                    XML=query_xml,
                    MaxHits=max_hits,
                    PageSize=page_size,
                    Offset=offset,
                    DateFormat='%Y-%m-%d', # %z,
                    FileInfo=1 if include_file_info else 0,
                    FollowDocLink=1 if follow_doc_links else 0,
                    Status=1 if include_status else 0,
                    GarbageMode=1 if garbage_mode else 0,
                )
        if context_user:
            job.append(Param('$$$SwitchContextUserName$$$', ParamTypes.STRING, context_user))

        current_offset = offset
        combined_count = 0
        columns = None
        while (not max_hits) or (combined_count <= max_hits):
            result = self.execute(job)
            if result.return_code != 0:
                raise RuntimeError(f'enaio error {result.return_code}: {result.error_message}')
            combined_count += result.values['Count']
            if result.values['TotalHits'] == 0:
                logging.debug(f'Query returned no result objects. Finished.')
                return

            archive = XmlElement.from_string(result.values['XML'])['Archive'][0]
            objecttype = archive['ObjectType'][0]
            rowset = objecttype['Rowset'][0]

            if not columns:
                columns = { column.attributes['internal_name']: column.attributes['datatype'] for column in rowset['Columns'][0]['Column'] }

            for row in rowset['Rows'][0]['Row']:
                yield_result = {}

                #Add common fields form Archive and ObjectType tag
                yield_result['OBJECT_FOLDER_TYPE_ID'] = archive.attributes['id']
                yield_result['OBJECT_TYPE_ID'] = objecttype.attributes['id']
                yield_result['OBJECT_MAINTYPE'] = objecttype.attributes['maintype']
                yield_result['OBJECT_COTYPE'] = objecttype.attributes['cotype']
                yield_result['OBJECT_INTERNAL_NAME'] = objecttype.attributes['internal_name']
                yield_result['OBJECT_TYPE'] = objecttype.attributes['type']
                yield_result['OBJECT_TABLE'] = objecttype.attributes['table']

                for name, value in zip(columns, row['Value']):
                    datatype = columns[name]
                    if datatype == 'TEXT':
                        # Value attribute of system fields or text value
                        yield_result[name] = value.attributes['value'] if 'value' in value.attributes else value.text
                    elif datatype == 'INTEGER':
                        # Text value of status fields or integer value of text value or None
                        yield_result[name] = value.text if 'value' in value.attributes else int(value.text) if value.text else None
                    elif datatype == 'DECIMAL': 
                        yield_result[name] = float(value.text) if value.text else None
                    elif datatype == 'DATE': 
                        yield_result[name] = datetime.strptime(value.text, '%Y-%m-%d').date() if value.text else None
                    elif datatype == 'DATETIME': 
                        yield_result[name] = datetime.strptime(value.text, '%Y-%m-%d %H:%M:%S') if value.text else None
                    else:
                        yield_result[name] = value.text
                    
                yield yield_result

            if page_size > result.values['Count']:
                logging.debug(f'Last query reached end of result rows. Finished')
                return

            if max_hits and (max_hits - combined_count == 0):
                logging.debug(f'Max hits reached. Finished')
                return 
            elif max_hits and (max_hits - combined_count < page_size):
                job.update(Param('PageSize', ParamTypes.INTEGER, max_hits - combined_count))

            current_offset += page_size
            job.update(Param('Offset', ParamTypes.INTEGER, current_offset))
            logging.debug(f'Paging to next result frame with offset {current_offset} after {combined_count} combined hits.')



from protlib import CStruct, CInt, CArray
from .job_parameter_description import JobParameterDescription
from .job_parameter_data import JobParameterData

class JobParameters(CStruct):
    length = CInt()
    count = CInt()
    description = CArray(length="count", ctype=JobParameterDescription)
    data = CArray(length="count", ctype=JobParameterData)

    def get(self, name): 
        for data in self.data:
            dName = data.name.decode('ascii')
            if dName == name:
                return data.value
        
        return None


from __future__ import annotations
from datetime import datetime
from typing import Any
from XmlElement import XmlElement

from .const import ParamTypes

class Param():
    def __init__(self, name:str, type:ParamTypes, value:Any):
        """Create a new Param() object.

        Keyword arguments:
        name -- String with parameter name, i. e. 'Flags'
        type -- ParamTypes Enum element definiting the parameter type, i. e. ParamTypes.INTEGER
        value -- The value to store for this Param.
        """        
        self.name = name
        self.type = type
        self.value = value

    
    def __repr__(self) -> str:
        return f'{self.name}:{self.type.value} = {self.value}'


    @staticmethod
    def infer_type(name:str, value:Any) -> Param:
        """Factory function to create new Param() objects without explicitly 
        stating a ParamTypes enum element as the parameter type.
        The type is infered by checking the values type against int, bool, float, datetime 
        or if the value is a instance of XmlElement or has the name 'XML'.

        Keyword arguments:
        name -- String with parameter name, i. e. 'Flags'
        value -- The value to store for this Param.
        """
        if isinstance(value, bool):
            return Param(name, ParamTypes.BOOLEAN, value) 
        elif isinstance(value, int):
            return Param(name, ParamTypes.INTEGER, value)
        elif isinstance(value, float):
            return Param(name, ParamTypes.DOUBLE, value)
        elif isinstance(value, datetime):
            return Param(name, ParamTypes.DATE_TIME, value)
        elif isinstance(value, XmlElement):
            return Param(name, ParamTypes.BASE64, value.to_string())
        elif name == "XML":
            return Param(name, ParamTypes.BASE64, value)
        else:
            return Param(name, ParamTypes.STRING, value)


# from protlib import 
import socket
import ssl
import pathlib
import hashlib
import base64
import os
from struct import *
from uuid import uuid4
from datetime import datetime
from typing import List, Optional
from tempfile import gettempdir
from ..param import Param
from ..result import Result
from ..result_file import ResultFile
from ..result_file import ResultFileType
from ..const import ParamTypes
from .call_job_parameters import CallJobParameters
from .job_header import JobHeader
from .file_header import FileHeader
from .file_footer import FileFooter
from .response_job_parameters import ResponseJobParameters
from .job_parameters import JobParameters
from .job_parameter_description import JobParameterDescription
from .job_parameter_data import JobParameterData


class JobCaller:
    CERT = ("-----BEGIN CERTIFICATE-----\n"
        "MIIF/TCCA+WgAwIBAgIBFzANBgkqhkiG9w0BAQsFADCBgzELMAkGA1UEBhMCREUx"
        "DzANBgNVBAgMBkJlcmxpbjEPMA0GA1UEBwwGQmVybGluMRgwFgYDVQQKDA9PUFRJ"
        "TUFMIFNZU1RFTVMxFTATBgNVBAsMDEJsdWUgQ2VudHJhbDEhMB8GA1UEAwwYQmx1"
        "ZSBDZW50cmFsIGltZXJtZWRpYXRlMB4XDTE5MDEyODE3NDIwNFoXDTI3MDQxNjE3"
        "NDIwNFowdDELMAkGA1UEBhMCREUxDzANBgNVBAgMBkJlcmxpbjEPMA0GA1UEBwwG"
        "QmVybGluMRgwFgYDVQQKDA9PUFRJTUFMIFNZU1RFTVMxFTATBgNVBAsMDEJsdWUg"
        "Q2VudHJhbDESMBAGA1UEAwwJQXBwU2VydmVyMIICIjANBgkqhkiG9w0BAQEFAAOC"
        "Ag8AMIICCgKCAgEAxXSPc72aZr1wuLfpZ6Lpu7XQuS4EPImNQccJg+JUxRFhBQa1"
        "Dme+dYLpdA7t5L2f5o00AOMVu5sCncIQ66yxtQe/ZlCUPhL9bBWfj/Yfs89PFQKQ"
        "wCISfdBmtqsY6pEK89qelbnEtCWIjBEzAm5z9p6AOk9iBZvzKSQb9ehNzP2AMtP5"
        "k53b8dXhjKTNM92MNzvO92ETqKUMdauzopSklRVLk6T3DRgJs+wpoKUCVvtXyoF/"
        "feilzPabku0+Tomui8hjrXnlcHo94Gp7dVxMPsxwyzKbYLRAM6pQ4vhzv+QmAOwS"
        "qgP79MupckGf70DMenHDjntrSw4l10if4q19xo+FJU+lfbGdPOSEMM0M3ttaAZUl"
        "+5HltzmTxhg6djhZV+myLKHMTlPaW76MWlzj2nKSJ63mOb4NF22NSuWSE8X2RRvO"
        "PNOZ3qaIV9a+2ljjPu8Sqmbo/ut7wcrty5ETCp3qZppM2vg8R4JfseTk5Nupfktf"
        "/lthfDMlX79VCou229yaLxAGqJv/WvXiy/zrGPLIrIsiiYv8aeAbxWK+hF3ea90B"
        "byHwKBAKT55RJ8fEqYWkxUdVEO5429UWKiZ0QhWas8NaZ2IokH2wjHfAKjHgVCBX"
        "mXU37lc6g3gNEL1NDPq5x9IxC+pWEx/H094xZ5AgyH4KF8h14uuczYrA8uECAwEA"
        "AaOBiTCBhjAdBgNVHQ4EFgQURyhC1GDxZgQGq+aT8dU15rVmZSYwHwYDVR0jBBgw"
        "FoAUPfCYO8wLyAv/Zl3XOrmZhx9cniswCQYDVR0TBAIwADALBgNVHQ8EBAMCBaAw"
        "LAYJYIZIAYb4QgENBB8WHU9wZW5TU0wgR2VuZXJhdGVkIENlcnRpZmljYXRlMA0G"
        "CSqGSIb3DQEBCwUAA4ICAQBwYbo4QKNwzV8Wy3mj9iUMHPp+fb/b1YTU7+wmqCmv"
        "ymyr83pQiuth2nu3y+7QEqHwDX+KJBm3XO1ej7GPZ5tZ6ssOsYQVGA5l7ujrsm/r"
        "aBr3n5cghiuggX2K9lbP8/I9HwvWPOqdtSHqy5ILQgQGR4mnh8zc1zPwIeYSLkeJ"
        "FjTChyx6ZX1p+gKYg+QQ4OzfLDbueTW/4oQrrH+DjvfD9yZnh0DRiL700NpQZxb4"
        "mnSqVS0rjMyLcNxc9M6IcivOqDy490CLcAj1KYjFP7B/Ehf4Po96p+geqDZRRs0v"
        "GfcRG4qaPi6mJ0p5Yf4PWeLN8ZBydg08pf3F4EDKcV+zzWoMq6ywwkUVPe5x+czJ"
        "cwKMIvCogaQRqzEBAuDGEMPkfz/Nl0wEy3zhx8gtVRvv/sjfCyyVm5rAS+ROkVAi"
        "w5njrOhKAGBYnVOfBEiCukCQYXNOP+/Rdi+J4QK81olPFnVpb07ltFhw68Gc+MXZ"
        "QKmed6+PTJak8/Wgqe/7SZtXq8NPElaLax7rfIQIrEgB01ow2PqhHdkQ8w0VKL2Z"
        "wSHyqIw00O/27DxxV1KWZkkNoen9RzY5YzzIknJ+4IaP18hyOfHm4bjrzUGvYumG"
        "WumfghHwh0F3EQEqH/T8vlGOALteuzG8aSNs9CWqtOK5qvQn+B9eUmjldwrVI0hR"
        "NA==\n-----END CERTIFICATE-----")

    def __init__(self, hostname, port, use_ssl: Optional[bool] = True, file_cache_byte_limit: Optional[int] = 33554432):
        self.hostname = hostname
        self.port = port
        self.file_cache_byte_limit = file_cache_byte_limit
        plain_text_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        plain_text_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        if use_ssl:
            ssl_context = ssl.SSLContext()
            ssl_context.load_verify_locations(cadata=self.CERT)

            self.socket = ssl_context.wrap_socket(ssl.wrap_socket(
                plain_text_socket, ssl_version=ssl.PROTOCOL_TLSv1, ciphers="AES256-SHA"))
        else: 
            self.socket = plain_text_socket
        self.socket.connect((hostname, port))

    def close(self):
        self.socket.close()

    def execute(self, job):
        # Create internal parameter
        intern_params: List[Param] = []
        intern_params.append(Param("streams", ParamTypes.INTEGER, len(job.files)))
        internal_parameters = self.build_parameters(intern_params)

        # create parameter
        parameters = self.build_parameters(job.params)
       
        # create request parameters 
        request_parameters = CallJobParameters(method=job.name, internal_parameters=internal_parameters, parameters=parameters)
        request_parameters_binary = request_parameters.serialize()
        
        # calculate digest of request parameters
        request_digest = hashlib.sha1()
        request_digest.update(request_parameters_binary)

        # create request headers
        request_headers = JobHeader()
        request_headers.set_parameter_length(len(request_parameters_binary) + 20)
        request_header_binary = request_headers.serialize()

        # send request
        self.socket.send(request_header_binary)
        self.socket.send(request_parameters_binary)
        
        # for each file
        for file in job.files:
            # read file info
            with pathlib.Path(file) as p: 
                size = p.stat().st_size 
                extension = p.suffix[1:]

                # write file header 
                request_file_header = FileHeader()
                request_file_header.set_file_length(size)
                request_file_header.extension = extension
                request_file_header_binary = request_file_header.serialize()
            
                self.socket.send(request_file_header_binary)
                request_digest.update(request_file_header_binary)

                # stream file
                f = open(file, 'rb')
                l = f.read(1024)
                while (l):
                    self.socket.send(l)
                    request_digest.update(l)
                    l = f.read(1024)
                f.close()

                # write fix file footer @0000000000@MAERTSSA
                request_file_footer = FileFooter()
                request_file_footer_binary = request_file_footer.serialize()
                self.socket.send(request_file_footer_binary)

                request_digest.update(request_file_footer_binary)

        # send request digest
        self.socket.send(request_digest.digest())

        # ----------------------------------------------------------------------
        # Start read response
        # ----------------------------------------------------------------------
        
        # read header
        response_header = JobHeader.parse(self.socket.recv(20))

        # read body
        response_digest = hashlib.sha1()
        
        parameter_length = response_header.get_parameter_length() - 20
        body_raw = b''
        while (len(body_raw) < parameter_length):
            body_raw += self.socket.recv(parameter_length - len(body_raw))

        response_digest.update(body_raw)

        response = self.parse_response_body(body_raw)
        
        return_code = response["internal_parameters"]["return"]
        error_message = None
        if return_code != 0:
            error_message = ''
            errors = response["errors"]
            for error in errors:
                error_message += error["message"] + '\n'

        result_values = response["parameters"]
        param_index = 0
        
        # parse file streams

        result_files: List[ResultFile] = []

        if 'streams' in response["internal_parameters"]:
            streams = response["internal_parameters"]["streams"]
            for i in range(0, int(streams)):
                header_raw = self.socket.recv(32)
                header = FileHeader.parse(header_raw)
                response_digest.update(header_raw)

                file_length = header.get_file_length()
                to_file = file_length >= self.file_cache_byte_limit
                remainder = file_length
                file_pointer = None
                byte_array = None
                file_path = None
                file_name = f'ecmind_{str(uuid4())}.{header.get_file_extension()}'
                if to_file:
                    file_path = os.path.join(gettempdir(), file_name)
                    file_pointer = open(file_path, 'wb')
                else:
                    byte_array = bytearray()
                BUFFER_SIZE = 4096
                while remainder > 0:
                    file_part:bytes = self.socket.recv(min(remainder, BUFFER_SIZE))
                    response_digest.update(file_part)
                    remainder -= len(file_part)
                    if to_file:
                        file_pointer.write(file_part)
                    else:
                        byte_array += bytearray(file_part)

                if to_file:
                    file_pointer.close()
                    rf = ResultFile(
                        result_file_type=ResultFileType.FILE_PATH,
                        file_name=file_name,
                        file_path=file_path
                    )
                else:
                    rf = ResultFile(
                        result_file_type=ResultFileType.BYTE_ARRAY,
                        file_name=file_name,                        
                        byte_array=bytes(byte_array)
                    )

                result_files.append(rf)
                
                footer_raw = self.socket.recv(20)
                response_digest.update(footer_raw)
                #footer = FileFooter.parse(footer_raw)

        response_digest_received = response_digest.digest()
        response_digest_expected = self.socket.recv(20)

        if(response_digest_received != response_digest_expected): 
            raise Exception("Digest for response does not match")

        return Result(result_values, result_files, return_code, error_message)

    def build_parameters(self, params: List[Param]):
        
        data_offset = 4 + (len(params) * 12)
        data_length = 0

        parameter_description_list = []
        parameter_description_size = 0

        parameter_data_list = []
        parameter_data_size = 0

        for param in params:
            parameter_description = JobParameterDescription(name_offset=data_offset + data_length, type=param.type.value, value_offset=data_offset + data_length + len(param.name) + 1)
            
            parameter_data = None
            if param.type == ParamTypes.STRING:
                parameter_data = JobParameterData(name=param.name, value=param.value)
            if param.type == ParamTypes.INTEGER:
                parameter_data = JobParameterData(name=param.name, value=str(param.value))
            if param.type == ParamTypes.BOOLEAN:
                parameter_data = JobParameterData(name=param.name, value="1" if param.value else "0")
            if param.type == ParamTypes.DOUBLE:
                parameter_data = JobParameterData(name=param.name, value=str(param.value))
            if param.type == ParamTypes.DATE_TIME:
                if isinstance(param.value, datetime):
                    JobParameterData(name=param.name, value=str(param.value.strftime("%d.%m.%y %H:%M:%S")))
                else:
                    JobParameterData(name=param.name, value=str(param.value))
            if param.type == ParamTypes.BASE64:
                if isinstance(param.value, (bytearray, bytes)):
                    parameter_data = JobParameterData(name=param.name, value=base64.b64encode(param.value))
                else:
                    parameter_data = JobParameterData(name=param.name, value=base64.b64encode(param.value.encode('UTF-8')))

            data_length += parameter_data.sizeof()
            
            parameter_description_list.append(parameter_description)
            parameter_description_size += parameter_description.sizeof() 

            parameter_data_list.append(parameter_data)
            parameter_data_size += parameter_data.sizeof()
        
        params_length = 4 + parameter_description_size + parameter_data_size
        
        return JobParameters(length=params_length, count=len(params), description=parameter_description_list, data=parameter_data_list)

    def parse_response_body(self, body_raw):
        result = {}

        # Read Mode
        mode = unpack("s", body_raw[0:1]) # Always R
        body_raw = body_raw[1:] # drop bytes
        
        # Read Internal Parameters
        internal_length = unpack(">i", body_raw[0:4])[0] # read byte length of internal parameters
        body_raw = body_raw[4:] # drop bytes

        internal_raw = body_raw[:internal_length] # slice internal parameters 
        body_raw = body_raw[internal_length:] # drop bytes

        result['internal_parameters'] = self.parse_parameters(internal_raw)

        # Read Parameters
        params_length = unpack(">i", body_raw[0:4])[0] # read byte length of parameters
        body_raw = body_raw[4:] # drop bytes
        
        params_raw = body_raw[:params_length] # slice parameters 
        body_raw = body_raw[params_length:] # drop bytes
        
        result['parameters'] = self.parse_parameters(params_raw)

        # Read Errors
        error_length = unpack(">i", body_raw[0:4])[0] # read byte length of errors
        error_raw = body_raw[4:] # drop bytes

        result['errors'] = self.parse_errors(error_raw)

        return result

    def parse_parameters(self, params_raw):
        parameters = {}
        # Read Parameter Count
        params_count = unpack(">i", params_raw[0:4])[0]

        for i in range(0, params_count):
            descriptionOffset = 4+(12 * i)
            name_offset = unpack(">i", params_raw[descriptionOffset: descriptionOffset + 4])[0]
            param_type = unpack(">i", params_raw[descriptionOffset+4: descriptionOffset + 8])[0]
            value_offset = unpack(">i", params_raw[descriptionOffset+8: descriptionOffset + 12])[0]

            value_offset_end = -1
            if(i == params_count - 1):
                value_offset_end = len(params_raw) - 1
            else:
                value_offset_end = unpack(">i", params_raw[descriptionOffset + 12:descriptionOffset + 16])[0] - 1

            name = str(params_raw[name_offset: value_offset-1], 'ascii')
            value = params_raw[value_offset: value_offset_end]

            infered_value = None
            if param_type == ParamTypes.STRING.value:
                infered_value = str(value, 'UTF-8')
            elif param_type == ParamTypes.INTEGER.value:
                infered_value = int(value)
            elif param_type == ParamTypes.BOOLEAN.value:
                infered_value = bool(value)
            elif param_type == ParamTypes.DOUBLE.value:
                infered_value = float(value)
            elif param_type == ParamTypes.DATE_TIME.value:
                infered_value = datetime(value)
            elif param_type == ParamTypes.BASE64.value:
                infered_value = base64.b64decode(value).decode('UTF-8')

            parameters[name] = infered_value

        return parameters

    def parse_errors(self, errors_raw):
        errors = []
        # Read Parameter Count
        error_count = unpack(">i", errors_raw[0:4])[0]
        errors_raw = errors_raw[4:] # drop bytes

        # dummy
        dummy = unpack(">i", errors_raw[0:4])[0]
        errors_raw = errors_raw[4:] # drop bytes

        descriptions = []

        for i in range(0, error_count):
            message_length, source_code, source_name_length, error_code, info_elements_length  = unpack(">iiiii", errors_raw[0: 20])
            errors_raw = errors_raw[20:] # drop bytes

            info_elements = []
            for c in range(0, info_elements_length):
                info_elements.append(unpack(">i", errors_raw[0:4])[0])
                errors_raw = errors_raw[4:] # drop bytes

            descriptions.append({
                "message_length": message_length, 
                "source_code": source_code, 
                "source_name_length": source_name_length, 
                "error_code": error_code,
                "info_elements_length": info_elements_length,
                "info_elements": info_elements
            })
        
        for description in descriptions:
            message = str(errors_raw[:description["message_length"]], "utf-8")
            errors_raw = errors_raw[description["message_length"]:] # drop bytes

            source = str(errors_raw[:description["source_name_length"]], "utf-8")
            errors_raw = errors_raw[description["source_name_length"]:] # drop bytes

            infos = []
            for info_element in description["info_elements"]:
                info = str(errors_raw[:info_element], "utf-8")
                infos.append(info)
                errors_raw = errors_raw[info_element:]
            
            errors.append({
                "message": message,
                "source_code": description["source_code"],
                "source": source,
                "error_code": description["error_code"],
                "infos": infos
            })
        
        return errors



class Options():
    def __init__(self,
            APPENDFILESTOFRONT:bool=None,
            ARCHIVABLE:bool=None,
            ARCHIVEIMMEDIATELY:bool=None,
            ARCHIVEIMMEDIATELYOBJDEF:bool=None,
            CHECKACCESS:bool=None,
            CHECKCATALOGUE:bool=None,
            CHECKEXISTENCE:bool=None,
            CHECKKEYFIELDS:bool=None,
            CHECKOBLIGATION:bool=None,
            CHECKPOSITION:bool=None,
            CHECKREADONLY:bool=None,
            DELETECASCADING:bool=None,
            COPYCASCADING:bool=None,
            FULLTEXTFILEATTACHED:bool=None,
            HARDDELETE:bool=None,
            INITFIELDS:bool=None,
            INUSERTRAY:bool=None,
            INWFTRAY:bool=None,
            REPLACEFILES:bool=None,
            REPLACEMULTIFIELDS:bool=None,
            REPLACEREMARKS:bool=None,
            REPLACETABLEFIELDS:bool=None,
            TRUNCATEVALUES:bool=None,
            TYPELESS:bool=None,
            UPDATEALLFIELDS:bool=None,
            VARIANTSAMELEVEL:bool=None,
            VARIANTSETACTIVE:bool=None,
            VARIANTTRANSFERRETENTION:bool=None,
            LINKDOCUMENT:bool=None,
            WFTOUSERTRAY:bool=None,
            KEEPLINKWHENEXISTS:bool=None,
            DELETEVARIANTMODE:bool=None,
            COPYINDEXONLY:bool=None,
            COPYCREATEHISTORY:bool=None
        ):
        self.APPENDFILESTOFRONT = APPENDFILESTOFRONT
        self.ARCHIVABLE = ARCHIVABLE
        self.ARCHIVEIMMEDIATELY = ARCHIVEIMMEDIATELY
        self.ARCHIVEIMMEDIATELYOBJDEF = ARCHIVEIMMEDIATELYOBJDEF
        self.CHECKACCESS = CHECKACCESS
        self.CHECKCATALOGUE = CHECKCATALOGUE
        self.CHECKEXISTENCE = CHECKEXISTENCE
        self.CHECKKEYFIELDS = CHECKKEYFIELDS
        self.CHECKOBLIGATION = CHECKOBLIGATION
        self.CHECKPOSITION = CHECKPOSITION
        self.CHECKREADONLY = CHECKREADONLY
        self.DELETECASCADING = DELETECASCADING
        self.COPYCASCADING = COPYCASCADING
        self.FULLTEXTFILEATTACHED = FULLTEXTFILEATTACHED
        self.HARDDELETE = HARDDELETE
        self.INITFIELDS = INITFIELDS
        self.INUSERTRAY = INUSERTRAY
        self.INWFTRAY = INWFTRAY
        self.REPLACEFILES = REPLACEFILES
        self.REPLACEMULTIFIELDS = REPLACEMULTIFIELDS
        self.REPLACEREMARKS = REPLACEREMARKS
        self.REPLACETABLEFIELDS = REPLACETABLEFIELDS
        self.TRUNCATEVALUES = TRUNCATEVALUES
        self.TYPELESS = TYPELESS
        self.UPDATEALLFIELDS = UPDATEALLFIELDS
        self.VARIANTSAMELEVEL = VARIANTSAMELEVEL
        self.VARIANTSETACTIVE = VARIANTSETACTIVE
        self.VARIANTTRANSFERRETENTION = VARIANTTRANSFERRETENTION
        self.LINKDOCUMENT = LINKDOCUMENT
        self.WFTOUSERTRAY = WFTOUSERTRAY
        self.KEEPLINKWHENEXISTS = KEEPLINKWHENEXISTS
        self.DELETEVARIANTMODE = DELETEVARIANTMODE
        self.COPYINDEXONLY = COPYINDEXONLY
        self.COPYCREATEHISTORY = COPYCREATEHISTORY
        return

    def __str__(self):
        result = ''
        if self.APPENDFILESTOFRONT != None:
            result += 'APPENDFILESTOFRONT=' + ('1' if self.APPENDFILESTOFRONT else '0') + ';'
        if self.ARCHIVABLE != None:
            result += 'ARCHIVABLE=' + ('1' if self.ARCHIVABLE else '0') + ';'
        if self.ARCHIVEIMMEDIATELY != None:
            result += 'ARCHIVEIMMEDIATELY=' + ('1' if self.ARCHIVEIMMEDIATELY else '0') + ';'
        if self.ARCHIVEIMMEDIATELYOBJDEF != None:
            result += 'ARCHIVEIMMEDIATELYOBJDEF=' + ('1' if self.ARCHIVEIMMEDIATELYOBJDEF else '0') + ';'
        if self.CHECKACCESS != None:
            result += 'CHECKACCESS=' + ('1' if self.CHECKACCESS else '0') + ';'
        if self.CHECKCATALOGUE != None:
            result += 'CHECKCATALOGUE=' + ('1' if self.CHECKCATALOGUE else '0') + ';'
        if self.CHECKEXISTENCE != None:
            result += 'CHECKEXISTENCE=' + ('1' if self.CHECKEXISTENCE else '0') + ';'
        if self.CHECKKEYFIELDS != None:
            result += 'CHECKKEYFIELDS=' + ('1' if self.CHECKKEYFIELDS else '0') + ';'
        if self.CHECKOBLIGATION != None:
            result += 'CHECKOBLIGATION=' + ('1' if self.CHECKOBLIGATION else '0') + ';'
        if self.CHECKPOSITION != None:
            result += 'CHECKPOSITION=' + ('1' if self.CHECKPOSITION else '0') + ';'
        if self.CHECKREADONLY != None:
            result += 'CHECKREADONLY=' + ('1' if self.CHECKREADONLY else '0') + ';'
        if self.DELETECASCADING != None:
            result += 'DELETECASCADING=' + ('1' if self.DELETECASCADING else '0') + ';'
        if self.COPYCASCADING != None:
            result += 'COPYCASCADING=' + ('1' if self.COPYCASCADING else '0') + ';'
        if self.FULLTEXTFILEATTACHED != None:
            result += 'FULLTEXTFILEATTACHED=' + ('1' if self.FULLTEXTFILEATTACHED else '0') + ';'
        if self.HARDDELETE != None:
            result += 'HARDDELETE=' + ('1' if self.HARDDELETE else '0') + ';'
        if self.INITFIELDS != None:
            result += 'INITFIELDS=' + ('1' if self.INITFIELDS else '0') + ';'
        if self.INUSERTRAY != None:
            result += 'INUSERTRAY=' + ('1' if self.INUSERTRAY else '0') + ';'
        if self.INWFTRAY != None:
            result += 'INWFTRAY=' + ('1' if self.INWFTRAY else '0') + ';'
        if self.REPLACEFILES != None:
            result += 'REPLACEFILES=' + ('1' if self.REPLACEFILES else '0') + ';'
        if self.REPLACEMULTIFIELDS != None:
            result += 'REPLACEMULTIFIELDS=' + ('1' if self.REPLACEMULTIFIELDS else '0') + ';'
        if self.REPLACEREMARKS != None:
            result += 'REPLACEREMARKS=' + ('1' if self.REPLACEREMARKS else '0') + ';'
        if self.REPLACETABLEFIELDS != None:
            result += 'REPLACETABLEFIELDS=' + ('1' if self.REPLACETABLEFIELDS else '0') + ';'
        if self.TRUNCATEVALUES != None:
            result += 'TRUNCATEVALUES=' + ('1' if self.TRUNCATEVALUES else '0') + ';'
        if self.TYPELESS != None:
            result += 'TYPELESS=' + ('1' if self.TYPELESS else '0') + ';'
        if self.UPDATEALLFIELDS != None:
            result += 'UPDATEALLFIELDS=' + ('1' if self.UPDATEALLFIELDS else '0') + ';'
        if self.VARIANTSAMELEVEL != None:
            result += 'VARIANTSAMELEVEL=' + ('1' if self.VARIANTSAMELEVEL else '0') + ';'
        if self.VARIANTSETACTIVE != None:
            result += 'VARIANTSETACTIVE=' + ('1' if self.VARIANTSETACTIVE else '0') + ';'
        if self.VARIANTTRANSFERRETENTION != None:
            result += 'VARIANTTRANSFERRETENTION=' + ('1' if self.VARIANTTRANSFERRETENTION else '0') + ';'
        if self.LINKDOCUMENT != None:
            result += 'LINKDOCUMENT=' + ('1' if self.LINKDOCUMENT else '0') + ';'
        if self.WFTOUSERTRAY != None:
            result += 'WFTOUSERTRAY=' + ('1' if self.WFTOUSERTRAY else '0') + ';'
        if self.KEEPLINKWHENEXISTS != None:
            result += 'KEEPLINKWHENEXISTS=' + ('1' if self.KEEPLINKWHENEXISTS else '0') + ';'
        if self.DELETEVARIANTMODE != None:
            result += 'DELETEVARIANTMODE=' + ('1' if self.DELETEVARIANTMODE else '0') + ';'
        if self.COPYINDEXONLY != None:
            result += 'COPYINDEXONLY=' + ('1' if self.COPYINDEXONLY else '0') + ';'
        if self.COPYCREATEHISTORY != None:
            result += 'COPYCREATEHISTORY=' + ('1' if self.COPYCREATEHISTORY else '0') + ';'

        return result.rstrip(';')

    __repr__ = __str__


from protlib import CStruct, CString
from .job_parameters import JobParameters
from .response_job_errors import ResponseJobErrors

class ResponseJobParameters(CStruct):
    mode = CString(default="R", length=1)  # always R for response
    internal_parameters = JobParameters.get_type()
    parameters = JobParameters.get_type()
    errors = ResponseJobErrors.get_type()


from typing import List, Union
from .const import FieldGroupOperators, SortOrder, SystemFields, QueryOperators, SpecialValues
from XmlElement import XmlElement
from datetime import datetime, date

class QueryConditionField():
    def __init__(
            self, 
            field:Union[str, SystemFields], 
            values:Union[str,int,float,date,datetime,None,SpecialValues, List[Union[str,int,float,date,datetime,None,SpecialValues]]],
            operator:QueryOperators=QueryOperators.EQUAL,
            table:str=None
        ):
        """Create a new QueryConditionField() object.

        Keyword arguments:
        field -- A internal name as string or a instance of SystemFields. Internal names beeing automatically checked against known system field names.
        operator -- (Optional) QueryOperators instance, default = EQUAL.
        table -- (Optional) Table internal name. If this is set, the a `<TableCondition>` is created using the field parameter as `<TableColumn>`, default = None.
        values: A list of literal values, None (interpretet as `<NULL />` value) or a SpecialValues instance for `<SpecialValues>`
        """ 

        if not table and isinstance(field, SystemFields):
            self.internal_name = field.name
            self.system = True
        elif not table and field in [ system_field.name for system_field in SystemFields]:
            self.internal_name = field
            self.system = True
        else:
            self.internal_name = field
            self.system = False

        self.query_operator = operator
        self.table_name = table
        self.values = values

    def __repr__(self) -> str:
        return (f'{self.table_name}:' if self.table_name else '') + f'{self.internal_name} {self.query_operator.value} {self.values}'

    def to_xml_element(self) -> XmlElement:
        """Render self to XmlElement"""

        def values_to_xml_elements(self) -> List[XmlElement]:
            if not self.values:
                return [ XmlElement('NULL') ]
            result = []
            for v in self.values if isinstance(self.values, list) else [ self.values ]:
                if v == None:
                    result.append(XmlElement('NULL'))
                elif isinstance(v, SpecialValues):
                    result.append(XmlElement('SpecialValue', t=v.value))
                elif isinstance(v, datetime):
                    result.append(XmlElement('Value', t=datetime.strftime(v, '%d.%m.%Y %H:%M:%S')))
                elif isinstance(v, date):
                    v_datetime = datetime.combine(v, datetime.min.time())
                    result.append(XmlElement('Value', t=datetime.strftime(v_datetime, '%d.%m.%Y')))
                else:
                    result.append(XmlElement('Value', t=str(v)))
            
            if len(result) == 0:
                return [ XmlElement('NULL') ]
            
            return result

        if self.table_name:
            x = XmlElement('TableCondition', {'internal_name': self.table_name}, [
                XmlElement(
                    'TableColumn', {
                        'internal_name': self.internal_name, 
                        'operator': self.query_operator.value 
                    },
                    values_to_xml_elements(self)
                )
            ])
        else:
            x = XmlElement(
                    'FieldCondition', {
                        'internal_name': self.internal_name,
                        'operator': self.query_operator.value 
                    },
                    values_to_xml_elements(self)
                )
            if self.system:
                x.attributes['system'] = '1'
        
        return x


from typing import List, Union
from .const import FieldGroupOperators, SortOrder, SystemFields
from .query_condition_field import QueryConditionField
from XmlElement import XmlElement

class QueryConditionGroup():
    def __init__(
            self, 
            object_name:str,
            field_conditions:Union[QueryConditionField, List[QueryConditionField]],
            group_operator:FieldGroupOperators=FieldGroupOperators.AND,
        ):
        """Create a new QueryConditionGroup() object combining the `<ConditionObject>` and `<FieldGroup>` tags.

        TODO
        Keyword arguments:
        object_name -- The internal name of a folder, register or document type.
        field_conditions -- A list of or single QueryConditionField(s).
        group_operator -- (Optional) FieldGroupOperators, default = FieldGroupOperators.AND.
        """ 
        self.internal_name = object_name
        if isinstance(field_conditions, list):
            self.field_conditions = field_conditions
        else:
            self.field_conditions = [ field_conditions ]
        
        self.group_operator = group_operator


    def __repr__(self) -> str:
        return f'{self.internal_name} ({self.group_operator.value}): {self.field_conditions}'

    
    def to_xml_element(self) -> XmlElement:
        """Render self to XmlElement"""
        return XmlElement(
            'ConditionObject', 
            {'internal_name': self.internal_name},
            [ x.to_xml_element() for x in self.field_conditions ]
        )


from .result_file import ResultFile
from typing import List, Optional

class Result():
    def __init__(self, values:dict={}, files:List[ResultFile]=[], return_code:int=-1, error_message:Optional[str]='', client_infos:Optional[dict]={}):
        """Create a new Result() object.

        Keyword arguments:
        values -- Dictionary of output parameters.
        files -- List of ResultFile() output file parameters.
        return_code -- Integer representation of the job result.
        error_message -- String containing error responses from the server on None if return_code is 0.
        client_infos -- List of optional data returned by the individual client implementation.
        """
        self.values = values
        self.files = files
        self.return_code = return_code
        self.error_message = error_message
        self.client_infos = client_infos

    def __repr__(self):
        if self.return_code == 0:
            return f'Result (success, {len(self.files)} files): {self.values}'
        else:
            return f'Result (failed, code {self.return_code}): {self.error_message}'


from protlib import CStruct, CString, AUTOSIZED
from .job_parameters import JobParameters

class CallJobParameters(CStruct):
    mode = CString(default="C", length=1)  # always C for call
    method = CString(length=AUTOSIZED)
    internal_parameters = JobParameters.get_type()
    parameters = JobParameters.get_type() 


from protlib import CStruct, CString, AUTOSIZED

class JobParameterData(CStruct):
    name = CString(length=AUTOSIZED)
    value = CString(length=AUTOSIZED)

    def sizeof(self) -> int:
        return len(self.name) + len(self.value) + 2


from protlib import CStruct, CString

class JobHeader(CStruct):
    dummy1 = CString(default="L:", length=2)  # always L-
    protocol = CString(default="BIN", length=3)  # BIN or XML
    dummy3 = CString(default="-", length=1)  # always -
    parameter_length = CString(length=10)  # Zero filed byte length of parameters
    version = CString(default="v50", length=3)  # always v50
    compression = CString(default="N", length=1)  # Y or N
    
    def set_parameter_length(self, length):
        self.parameter_length = f"{length:0>10}"

    def get_parameter_length(self):
        return int(self.parameter_length)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='ecmind_blue_client',
    version='0.4.1',
    author='Ulrich Wohlfeil, Roland Koller',
    author_email='info@ecmind.ch',
    description='A client wrapper for blue',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.ecmind.ch/open/ecmind_blue_client',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
    install_requires=['XmlElement>=0.3.0'],
    extras_require = {
        'SoapClient': ['zeep'],
        'ComClient': ['comtypes'],
        'TcpClient': ['protlib']
    }
)


import logging
import socket
from random import randint
from typing import Optional

from .client import Client
from .job import Job
from .result import Result
from .tcp_client_classes.job_caller import JobCaller

class TcpClient(Client):
    @staticmethod
    def encrypt_password(password:str) -> str:
        if password == None or password == '':
            raise(ValueError('Password must not be empty'))
        
        create_random_char = lambda: chr(ord('0') + (randint(0, 32000) % 8))

        plen = len(password)
        nmax = 4 * plen * 2
        ioff = randint(0, 32000) % max(1, nmax - plen * 3 - 3)
        cryptid = chr(ord('A') + plen) + chr(ord('A') + ioff)
        for _ in range(0, ioff):
            cryptid += create_random_char()

        replace_in_string = lambda i, c: cryptid[:i] + c + cryptid[i + 1:]

        for i in range(0, plen):
            j = 2 + ioff + i * 3
            oct_part = ioff + ord(password[i])
            oct = f'{oct_part:03o}'
            for k in range(0, 3):
                cryptid = replace_in_string(j+k, oct[k])

        for i in range(2 + ioff + 3 * plen, nmax):
            cryptid = replace_in_string(i, create_random_char())

        try:
            cryptid.encode('ascii')
            return cryptid
        except:
            return TcpClient.encrypt_password(password)

    def __attach__(self, username:str, password:str):
            session_attach_job = Job('krn.SessionAttach', Flags=0, SessionGUID='')
            session_attach_result = self.execute(session_attach_job)
            self.session_guid = session_attach_result.values['SessionGUID']

            session_properties_set_job = Job(
                'krn.SessionPropertiesSet', 
                Flags=0, 
                Properties='instname;statname;address', 
                address=f'{socket.gethostbyname(socket.gethostname())}=dummy',
                instname=self.appname,
                statname=socket.gethostname()
            )
            session_properties_set_result = self.execute(session_properties_set_job)

            session_login_job = Job(
                'krn.SessionLogin', 
                Flags=0, 
                UserName=username, 
                UserPwd=TcpClient.encrypt_password(password)
            )
            session_login_result = self.execute(session_login_job)

            if session_login_result.values['Description'] != None and session_login_result.values['Description'] != '':
                raise RuntimeError(f'Login error: {session_login_result.values["Description"]}')


    def __init__(self, 
            hostname:str, 
            port:int, 
            appname:str, 
            username:str, 
            password:str, 
            use_ssl:bool=True, 
            file_cache_byte_limit:int=33554432, 
            auto_reconnect:bool=True
        ):
        self.hostname = hostname
        self.port = port
        self.appname = appname
        self.username = username
        self.password = password
        self.use_ssl = use_ssl
        self.file_cache_byte_limit = file_cache_byte_limit
        self.auto_reconnect = auto_reconnect
        self._connect()

    def _connect(self):
        if hasattr(self, 'job_caller') and self.job_caller != None:
            # try to close existing job_caller 
            try:
                self.job_caller.close()
            except Exception as e:
                logging.warning(e)
            # remove job_caller reference
            self.job_caller = None
        
        self.job_caller = JobCaller(self.hostname, self.port, self.use_ssl, self.file_cache_byte_limit)
        self.__attach__(self.username, self.password)
        
    def __del__(self): 
        try:
            self.job_caller.socket.close()
        except:
            pass

    def execute(self, job:Job) -> Result:
        """Send a job to the blue server (via TCP), execute it and return the response.

        Keyword arguments:
        job -- A previously created Job() object.
        """
        
        if self.auto_reconnect:
            if self.job_caller == None:
                # try to connect if current job_caller is None
                self._connect()
            
            try:
                return self.job_caller.execute(job)
            except ConnectionAbortedError as e:
                # fetch connection closed exceptions and try to reconnect and execute again
                logging.warning(e)
                self._connect()
                return self.execute(job)
        else: 
            return self.job_caller.execute(job)

class Connection():
    def __init__(self, 
        hostname:str, 
        port:int, 
        appname:str, 
        username:str, 
        password:str, 
        use_ssl:bool=True, 
        file_cache_byte_limit:int=33554432, 
        auto_reconnect:bool=True
    ):
        self.hostname = hostname
        self.port = port
        self.appname = appname
        self.username = username
        self.password = password
        self.use_ssl = use_ssl
        self.file_cache_byte_limit = file_cache_byte_limit
        self.auto_reconnect = auto_reconnect

    def __enter__(self):
        self.client = TcpClient(
            hostname=self.hostname,
            port=self.port,
            appname=self.appname,
            username=self.username,
            password=self.password,
            use_ssl=self.use_ssl,
            file_cache_byte_limit=self.file_cache_byte_limit,
            auto_reconnect=self.auto_reconnect
        )
        return self.client
    
    def __exit__(self, type, value, traceback):
        self.client.__del__()